#include "MemoryTreeWidgetBuilder.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/gui-plugins/ObjectPoseGui/TreeWidgetBuilder.h>


namespace armarx::armem
{

    void MemoryTreeWidgetBuilder::updateChildren(const armem::Memory& memory, QTreeWidget* tree)
    {
        updateChildren(std::map<string, const armem::Memory*>{{memory.name, &memory}}, tree);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const std::map<string, const armem::Memory*>& memories, QTreeWidget* tree)
    {
        MapTreeWidgetBuilder<std::string, const Memory*> builder;

        builder.setMakeItemFn([](const std::string&, const Memory* memory)
        {
            return new QTreeWidgetItem({ QString::fromStdString(memory->name), "Memory", QString::number(memory->coreSegments.size()) });
        });
        builder.setUpdateItemFn([this](const std::string&, const Memory* memory, QTreeWidgetItem* memoryItem)
        {
            if (memoryItem)
            {
                updateChildren(*memory, memoryItem);
            }
            return true;
        });

        builder.updateTree(tree, memories);
    }



    void MemoryTreeWidgetBuilder::updateChildren(const armem::Memory& memory, QTreeWidgetItem* memoryItem)
    {
        MapTreeWidgetBuilder<std::string, CoreSegmentPtr> builder;
        builder.setMakeItemFn([](const std::string&, const CoreSegmentPtr& coreSeg)
        {
            return new QTreeWidgetItem({ QString::fromStdString(coreSeg->name), "Core Segment", QString::number(coreSeg->providerSegments.size()) });
        });
        builder.setUpdateItemFn([this](const std::string&, const CoreSegmentPtr& coreSeg, QTreeWidgetItem* coreSegItem)
        {
            if (coreSeg)
            {
                updateChildren(*coreSeg, coreSegItem);
            }
            return true;
        });

        builder.updateTree(memoryItem, memory.coreSegments);
    }


    void MemoryTreeWidgetBuilder::updateChildren(const armem::CoreSegment& coreSeg, QTreeWidgetItem* coreSegItem)
    {
        MapTreeWidgetBuilder<std::string, ProviderSegmentPtr> builder;
        builder.setMakeItemFn([](const std::string&, const ProviderSegmentPtr& provSeg)
        {
            return new QTreeWidgetItem({ QString::fromStdString(provSeg->name), "Provider Segment", QString::number(provSeg->entities.size()) });
        });
        builder.setUpdateItemFn([this](const std::string&, const ProviderSegmentPtr& provSeg, QTreeWidgetItem* provSegItem)
        {
            if (provSeg)
            {
                updateChildren(*provSeg, provSegItem);
            }
            return true;
        });

        builder.updateTree(coreSegItem, coreSeg.providerSegments);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::ProviderSegment& provSeg, QTreeWidgetItem* provSegItem)
    {
        MapTreeWidgetBuilder<std::string, EntityPtr> builder;
        builder.setMakeItemFn([](const std::string& name, const EntityPtr& entity)
        {
            (void) entity;
            return new QTreeWidgetItem({QString::fromStdString(name), "Entity", QString::number(entity->history.size())});
        });
        builder.setUpdateItemFn([this](const std::string&, const EntityPtr& entity, QTreeWidgetItem* entitySegItem)
        {
            if (entity)
            {
                updateChildren(*entity, entitySegItem);
            }
            return true;
        });

        builder.updateTree(provSegItem, provSeg.entities);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::Entity& entity, QTreeWidgetItem* entityItem)
    {
        MapTreeWidgetBuilder<armem::Time, EntitySnapshotPtr> builder;
        builder.setMakeItemFn([](const armem::Time& time, const EntitySnapshotPtr& snapshot)
        {
            (void) snapshot;
            return new QTreeWidgetItem({ QString::fromStdString(armem::toDateTimeMilliSeconds(time)), "Entity Snapshot", QString::number(snapshot->instances.size()) });
        });
        builder.setUpdateItemFn([this](const armem::Time&, const EntitySnapshotPtr& snapshot, QTreeWidgetItem* snapshotItem)
        {
            if (snapshot)
            {
                updateChildren(*snapshot, snapshotItem);
            }
            return true;
        });

        builder.updateTree(entityItem, entity.history);
    }

    void MemoryTreeWidgetBuilder::updateChildren(const armem::EntitySnapshot& snapshot, QTreeWidgetItem* snapshotItem)
    {
        TreeWidgetBuilder builder(snapshot.instances);
        builder.setMakeItemFn([](const EntityDataPtr& instance)
        {
            return new QTreeWidgetItem({ QString(instance->index) , "Entity Data" });
        });
        builder.setUpdateItemFn([](const EntityDataPtr& instance, QTreeWidgetItem* instanceItem)
        {
            if (instance)
            {
                (void) instance;
                (void) instanceItem;
            }
            return true;
        });

        builder.updateTree(snapshotItem, snapshot.instances);
    }

}
