#include "CoreSegment.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"


namespace armarx::armem
{

    CoreSegment::CoreSegment(const CoreSegment& other) : TypedEntityStorage(other)
    {
        *this = other;
    }

    CoreSegment& CoreSegment::operator=(const CoreSegment& other)
    {
        TypedEntityStorage::operator=(other);

        this->providerSegments.clear();
        for (const auto& [name, entity] : other.providerSegments)
        {
            this->providerSegments.emplace(name, std::make_unique<ProviderSegment>(*entity));
        }

        return *this;
    }

    CoreSegment::~CoreSegment()
    {
    }

    CoreSegmentPtr CoreSegment::getEmptyCopy() const
    {
        CoreSegmentPtr other = std::make_unique<CoreSegment>();
        other->name = this->name;
        return other;
    }

    bool CoreSegment::hasProviderSegment(const std::string& name) const
    {
        return providerSegments.count(name) > 0;
    }

    ProviderSegment& CoreSegment::getProviderSegment(const std::string& name)
    {
        return const_cast<ProviderSegment&>(const_cast<const CoreSegment*>(this)->getProviderSegment(name));
    }

    const ProviderSegment& CoreSegment::getProviderSegment(const std::string& name) const
    {
        auto it = this->providerSegments.find(name);
        if (it != providerSegments.end())
        {
            ARMARX_CHECK_NOT_NULL(it->second);
            return *it->second;
        }
        else
        {
            throw armem::error::MissingEntry("provider segment", name, getLevelName(), this->name);
        }
    }

    const Entity& CoreSegment::getEntity(const MemoryID& id) const
    {
        checkStorageName(id.coreSegmentName, this->name, getLevelName());
        return getProviderSegment(id.providerSegmentName).getEntity(id);
    }


    ProviderSegment& CoreSegment::addProviderSegment(const std::string& name, aron::type::AronTypePtr providerSegmentType)
    {
        aron::type::AronTypePtr type = providerSegmentType ? providerSegmentType : this->aronType;
        return addProviderSegment(std::make_unique<ProviderSegment>(name, type));
    }

    ProviderSegment& CoreSegment::addProviderSegment(const ProviderSegment& providerSegment)
    {
        return addProviderSegment(std::make_unique<ProviderSegment>(providerSegment));
    }

    ProviderSegment& CoreSegment::addProviderSegment(ProviderSegmentPtr&& providerSegment)
    {
        if (hasProviderSegment(name))
        {
            throw armem::error::StorageEntryAlreadyExists("provider segment", name, getLevelName(), this->name);
        }
        auto it = providerSegments.emplace(providerSegment->name, std::move(providerSegment)).first;
        ARMARX_CHECK_NOT_NULL(it->second);
        return *it->second;
    }


    MemoryID CoreSegment::update(const InternalEntityUpdate& update)
    {
        checkStorageName(update.entityID.coreSegmentName, this->name, getLevelName());

        auto it = this->providerSegments.find(update.entityID.providerSegmentName);
        if (it != providerSegments.end())
        {
            return it->second->update(update);
        }
        else
        {
            throw armem::error::MissingEntry("provider segment", update.entityID.providerSegmentName,
                                             getLevelName(), name);
        }
    }

}
