#include "Entity.h"

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error.h"


namespace armarx::armem
{

    Entity::Entity()
    {
    }

    Entity::Entity(const std::string& name) : name(name)
    {
    }

    Entity::Entity(const Entity& other)
    {
        *this = other;
    }

    Entity& Entity::operator=(const Entity& other)
    {
        this->name = other.name;
        this->maxHistorySize = other.maxHistorySize;

        this->history.clear();
        for (const auto& [time, snapshot] : other.history)
        {
            this->addSnapshot(*snapshot);
        }

        return *this;
    }

    EntityPtr Entity::getEmptyCopy() const
    {
        EntityPtr other = std::make_unique<Entity>();
        other->name = this->name;
        other->maxHistorySize = this->maxHistorySize;
        return other;
    }

    bool Entity::hasSnapshot(Time time) const
    {
        return history.count(time) > 0;
    }

    EntitySnapshot& Entity::getSnapshot(Time time)
    {
        return const_cast<EntitySnapshot&>(const_cast<const Entity*>(this)->getSnapshot(time));
    }

    const EntitySnapshot& Entity::getSnapshot(Time time) const
    {
        auto it = history.find(time);
        if (it != history.end())
        {
            ARMARX_CHECK_NOT_NULL(it->second);
            return *it->second;
        }
        else
        {
            throw error::MissingEntry("entity snapshot", toDateTimeMilliSeconds(time),
                                      getLevelName(), this->name);
        }
    }

    EntitySnapshot& Entity::getSnapshot(const MemoryID& id)
    {
        return const_cast<EntitySnapshot&>(const_cast<const Entity*>(this)->getSnapshot(id));
    }

    const EntitySnapshot& Entity::getSnapshot(const MemoryID& id) const
    {
        checkEntityName(id.entityName);
        return getSnapshot(id.timestamp);
    }

    EntitySnapshot& Entity::getLatestSnapshot()
    {
        return const_cast<EntitySnapshot&>(const_cast<const Entity*>(this)->getLatestSnapshot());
    }

    const EntitySnapshot& Entity::getLatestSnapshot() const
    {
        auto& latest = getLatestItem();
        ARMARX_CHECK_NOT_NULL(latest.second);
        return *latest.second;
    }

    Time Entity::getLatestTimestamp() const
    {
        return getLatestItem().first;
    }

    std::vector<Time> Entity::getTimestamps() const
    {
        return simox::get_keys(history);
    }


    EntitySnapshot& Entity::addSnapshot(const EntitySnapshot& snapshot)
    {
        return addSnapshot(std::make_unique<EntitySnapshot>(snapshot));
    }

    EntitySnapshot& Entity::addSnapshot(EntitySnapshotPtr&& snapshot)
    {
        auto it = history.emplace(snapshot->time, std::move(snapshot)).first;
        return *it->second;
    }


    MemoryID Entity::update(const InternalEntityUpdate& update)
    {
        checkEntityName(update.entityID.entityName);

        auto it = history.find(update.timeCreated);
        if (it == history.end())
        {
            // Insert into history.
            it = history.emplace(update.timeCreated, std::make_unique<EntitySnapshot>()).first;

            truncateHistoryToSize();
        }
        // else: Update snapshot

        // Update entry.
        ARMARX_CHECK_NOT_NULL(it->second);
        it->second->update(update);

        MemoryID snapshotID = update.entityID;
        snapshotID.timestamp = it->second->time;
        return snapshotID;
    }


    void Entity::setMaxHistorySize(long maxSize)
    {
        this->maxHistorySize = maxSize;
        truncateHistoryToSize();
    }



    void Entity::truncateHistoryToSize()
    {
        if (maxHistorySize >= 0)
        {
            while (history.size() > size_t(maxHistorySize))
            {
                history.erase(history.begin());
            }
            ARMARX_CHECK_LESS_EQUAL(history.size(), maxHistorySize);
        }
    }

    void Entity::checkEntityName(const std::string& name) const
    {
        if (name != this->name)
        {
            throw armem::error::StorageNameMismatch(name, getLevelName(), this->name);
        }
    }

    const std::map<Time, EntitySnapshotPtr>::value_type& Entity::getLatestItem() const
    {
        if (history.empty())
        {
            throw armem::error::EntityHistoryEmpty(name, "when getting the latest snapshot.");
        }
        return *history.rbegin();
    }

}
