#pragma once

#include <map>
#include <memory>
#include <string>

#include "../core/Time.h"

#include "EntitySnapshot.h"
#include "detail/MemoryContainer.h"


namespace armarx::armem
{

    class Entity;
    using EntityPtr = std::unique_ptr<Entity>;

    /**
     * @brief Data of an entity over a period of time.
     */
    class Entity :
        public detail::MemoryContainerImpl<std::map<Time, EntitySnapshotPtr>>
    {
    public:

        Entity();
        Entity(const std::string& name);

        /// Copy the history from `other` to this.
        Entity(const Entity& other);
        /// Copy the history from `other` to this.
        Entity& operator=(const Entity& other);


        bool hasSnapshot(Time time) const;

        EntitySnapshot& getSnapshot(Time time);
        const EntitySnapshot& getSnapshot(Time time) const;

        EntitySnapshot& getSnapshot(const MemoryID& id);
        const EntitySnapshot& getSnapshot(const MemoryID& id) const;

        /**
         * @brief Return the snapshot with the most recent timestamp.
         * @return The latest snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         */
        EntitySnapshot& getLatestSnapshot();
        const EntitySnapshot& getLatestSnapshot() const;

        /**
         * @brief Get the latest timestamp.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         */
        Time getLatestTimestamp() const;

        /**
         * @brief Get all timestamps in the history.
         */
        std::vector<Time> getTimestamps() const;


        /**
         * @brief Add the given update to this entity's history.
         * @param update The update.
         * @return The snapshot ID of the update.
         */
        MemoryID update(const InternalEntityUpdate& update);

        /// Add a single snapshot.
        EntitySnapshot& addSnapshot(const EntitySnapshot& snapshot);
        EntitySnapshot& addSnapshot(EntitySnapshotPtr&& snapshot);

        /// Get a cleared copy of `*this` (with empty history).
        EntityPtr getEmptyCopy() const;


        /**
         * @brief Sets the maximum history size.
         *
         * The current history is truncated if necessary.
         */
        void setMaxHistorySize(long maxSize);


    private:

        /// If maximum size is set, ensure `history`'s is not higher.
        void truncateHistoryToSize();
        void checkEntityName(const std::string& name) const;

        /**
         * @brief Return the snapshot with the most recent timestamp.
         * @return The latest snapshot.
         * @throw `armem::error::EntityHistoryEmpty` If the history is empty.
         */
        const std::map<Time, EntitySnapshotPtr>::value_type& getLatestItem() const;


    public:

        std::string name;
        std::map<Time, EntitySnapshotPtr> history;

        /**
         * @brief Maximum size of `history`
         *
         * If negative, the size of `history` is not limited.
         */
        long maxHistorySize = -1;
        // ToDo: Add max age;
        // ToDo in future: keep/remove predicate



        // MemoryLevel interface
    public:
        std::string getKeyString() const override
        {
            return name;
        }
        std::string getLevelName() const override
        {
            return "entity";
        }

        // MemoryContainer interface
    protected:
        ContainerT& _underlyingContainer() override
        {
            return history;
        }

    };

}
