#include "Memory.h"

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"


namespace armarx::armem
{

    Memory::Memory(const Memory& other) : EntityStorage(other)
    {
        *this = other;
    }

    Memory& Memory::operator=(const Memory& other)
    {
        this->name = other.name;

        this->coreSegments.clear();
        for (const auto& [name, coreSegment] : other.coreSegments)
        {
            this->coreSegments.emplace(name, std::make_unique<CoreSegment>(*coreSegment));
        }

        return *this;
    }

    Memory::~Memory()
    {
    }

    MemoryPtr Memory::getEmptyCopy() const
    {
        MemoryPtr other = std::make_unique<Memory>();
        other->name = this->name;
        return other;
    }

    bool Memory::hasCoreSegment(const std::__cxx11::string& name) const
    {
        return coreSegments.count(name) > 0;
    }

    CoreSegment& Memory::getCoreSegment(const std::string& name)
    {
        return const_cast<CoreSegment&>(const_cast<const Memory*>(this)->getCoreSegment(name));
    }
    const CoreSegment& Memory::getCoreSegment(const std::string& name) const
    {
        auto it = this->coreSegments.find(name);
        if (it != coreSegments.end())
        {
            ARMARX_CHECK_NOT_NULL(it->second);
            return *it->second;
        }
        else
        {
            throw armem::error::MissingEntry("core segment", name, getLevelName(), this->name);
        }
    }


    const Entity& Memory::getEntity(const MemoryID& id) const
    {
        checkStorageName(id.memoryName, this->name, getLevelName());
        return getCoreSegment(id.coreSegmentName).getEntity(id);
    }

    CoreSegment& Memory::addCoreSegment(const string& name, aron::type::AronTypePtr coreSegmentType)
    {
        return addCoreSegment(std::make_unique<CoreSegment>(name, coreSegmentType));
    }

    CoreSegment& Memory::addCoreSegment(const CoreSegment& coreSegment)
    {
        return addCoreSegment(std::make_unique<CoreSegment>(coreSegment));
    }

    CoreSegment& Memory::addCoreSegment(CoreSegmentPtr&& coreSegment)
    {
        if (coreSegments.count(name) > 0)
        {
            throw armem::error::StorageEntryAlreadyExists("core segment", name, getLevelName(), this->name);
        }
        auto it = coreSegments.emplace(coreSegment->name, std::move(coreSegment)).first;
        ARMARX_CHECK_NOT_NULL(it->second);
        return *it->second;
    }

    std::vector<CoreSegment*> Memory::addCoreSegments(const std::vector<std::string>& names)
    {
        std::vector<CoreSegment*> segments;
        for (const auto& name : names)
        {
            try
            {
                segments.push_back(&addCoreSegment(name));
            }
            catch (const armem::error::StorageEntryAlreadyExists& e)
            {
                ARMARX_INFO << e.what() << "\nIgnoring multiple addition.";
            }
        }
        return segments;
    }


    MemoryID Memory::update(const InternalEntityUpdate& update)
    {
        checkStorageName(update.entityID.memoryName, this->name, getLevelName());

        auto it = this->coreSegments.find(update.entityID.coreSegmentName);
        if (it != coreSegments.end())
        {
            return it->second->update(update);
        }
        else
        {
            throw armem::error::MissingEntry("core segment", update.entityID.coreSegmentName,
                                             getLevelName(), name);
        }
    }


}
