#include "ProviderSegment.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"


namespace armarx::armem
{

    ProviderSegment::ProviderSegment(const ProviderSegment& other) : TypedEntityStorage(other)
    {
        *this = other;
    }

    ProviderSegment& ProviderSegment::operator=(const ProviderSegment& other)
    {
        TypedEntityStorage::operator=(other);

        this->entities.clear();
        for (const auto& [name, entity] : other.entities)
        {
            this->entities.emplace(name, std::make_unique<Entity>(*entity));
        }

        return *this;
    }

    ProviderSegment::~ProviderSegment()
    {
    }

    ProviderSegmentPtr ProviderSegment::getEmptyCopy() const
    {
        ProviderSegmentPtr other = std::make_unique<ProviderSegment>();
        other->name = this->name;
        other->maxHistorySize = this->maxHistorySize;
        return other;
    }

    bool ProviderSegment::hasEntity(const string& name) const
    {
        return entities.count(name) > 0;
    }

    Entity& ProviderSegment::getEntity(const string& name)
    {
        return const_cast<Entity&>(const_cast<const ProviderSegment*>(this)->getEntity(name));
    }

    const Entity& ProviderSegment::getEntity(const std::string& name) const
    {
        auto it = this->entities.find(name);
        if (it != entities.end())
        {
            ARMARX_CHECK_NOT_NULL(it->second);
            return *it->second;
        }
        else
        {
            throw error::MissingEntry("entity", name, getLevelName(), this->name);
        }
    }

    const Entity& ProviderSegment::getEntity(const MemoryID& id) const
    {
        checkStorageName(id.providerSegmentName, this->name, getLevelName());
        return getEntity(id.entityName);
    }

    MemoryID ProviderSegment::update(const InternalEntityUpdate& update)
    {
        checkStorageName(update.entityID.providerSegmentName, this->name, getLevelName());

        auto it = this->entities.find(update.entityID.providerSegmentName);
        if (it == entities.end())
        {
            // Add entity entry.
            it = entities.emplace(update.entityID.entityName,
                                  std::make_unique<Entity>(update.entityID.entityName)).first;
            it->second->setMaxHistorySize(maxHistorySize);
        }
        // Update entity.
        return it->second->update(update);
    }

    Entity& ProviderSegment::addEntity(const string& name)
    {
        return addEntity(std::make_unique<Entity>(name));
    }

    Entity& ProviderSegment::addEntity(const Entity& entity)
    {
        return addEntity(std::make_unique<Entity>(entity));
    }

    Entity& ProviderSegment::addEntity(EntityPtr&& entity)
    {
        auto it = entities.emplace(entity->name, std::move(entity)).first;
        return *it->second;
    }

    void ProviderSegment::setMaxHistorySize(long maxSize)
    {
        this->maxHistorySize = maxSize;
        for (auto& [name, entity] : entities)
        {
            entity->setMaxHistorySize(maxSize);
        }
    }

}
