#pragma once

#include <map>
#include <memory>
#include <string>

#include "Entity.h"
#include "detail/EntityStorage.h"


namespace armarx::armem
{

    class ProviderSegment;
    using ProviderSegmentPtr = std::unique_ptr<ProviderSegment>;

    /**
     * @brief Data of a provider segment containing multiple entities.
     */
    class ProviderSegment : public detail::TypedEntityStorage<EntityPtr>
    {
    public:

        using TypedEntityStorage::TypedEntityStorage;

        ProviderSegment(const ProviderSegment& other);
        ProviderSegment& operator=(const ProviderSegment& other);

        ~ProviderSegment() override;


        bool hasEntity(const std::string& name) const;

        using TypedEntityStorage::getEntity;
        const Entity& getEntity(const MemoryID& id) const override;

        Entity& getEntity(const std::string& name);
        const Entity& getEntity(const std::string& name) const;


        using TypedEntityStorage::update;
        /**
         * @brief Updates an entity's history.
         *
         * Missing entity entries are added before updating.
         */
        virtual MemoryID update(const InternalEntityUpdate& update) override;

        /// Add an empty entity with the given name.
        Entity& addEntity(const std::string& name);
        /// Copy and insert an entity.
        Entity& addEntity(const Entity& entity);
        /// Move and insert an entity.
        Entity& addEntity(EntityPtr&& entity);

        /// Get an cleared copy of `*this` (without entities).
        ProviderSegmentPtr getEmptyCopy() const;


        /**
         * @brief Sets the maximum history size of entities in this segment.
         * This affects all current entities as well as new ones.
         * @see Entity::setMaxHistorySize()
         */
        void setMaxHistorySize(long maxSize);


    public:

        /// The entities.
        std::map<std::string, EntityPtr> entities;

        /**
         * @brief Maximum size of entity histories.
         * @see Entity::maxHstorySize
         */
        long maxHistorySize = -1;


        // MemoryLevel interface
    public:
        std::string getLevelName() const override
        {
            return "provider segment";
        }

        // MemoryContainer interface
    protected:
        ContainerT& _underlyingContainer() override
        {
            return entities;
        }

    };

}
