#pragma once

#include "../InternalCommit.h"
#include "../Entity.h"
#include "../EntitySnapshot.h"
#include "../../error/ArMemError.h"

#include "MemoryContainer.h"


namespace armarx::armem::detail
{

    /**
     * @brief A container of entities at some point in the hierarchy.
     *
     * Can be updated by multiple entity updates.
     */
    template <class ValueT>
    class EntityStorage : public MemoryContainerImpl<std::map<std::string, ValueT>>
    {
    public:

        EntityStorage()
        {
        }
        EntityStorage(const std::string& name) : name(name)
        {
        }

        virtual ~EntityStorage() override
        {
        }


        /**
         * @brief Store all updates in `commit`.
         * @param commit The commit.
         * @return The resulting memory IDs.
         */
        std::vector<MemoryID> update(const InternalCommit& commit)
        {
            std::vector<MemoryID> ids;
            for (const auto& update : commit.updates)
            {
                ids.push_back(this->update(update));
            }
            return ids;
        }
        /**
         * @brief Store the given update.
         * @param update The update.
         * @return The resulting entity snapshot's ID.
         */
        virtual MemoryID update(const InternalEntityUpdate& update) = 0;


        /**
         * @brief Retrieve an entity.
         * @param id The entity ID.
         * @return The entity.
         * @throw An exception deriving from `armem::error::ArMemError` if the entity is missing.
         */
        virtual Entity& getEntity(const MemoryID& id)
        {
            return const_cast<Entity&>(const_cast<const EntityStorage*>(this)->getEntity(id));
        }
        virtual const Entity& getEntity(const MemoryID& id) const = 0;

        /**
         * @brief Retrieve an entity snapshot.
         *
         * Uses `getEntity()` to retrieve the respective entity.
         *
         * @param id The snapshot ID.
         * @return The entity snapshot.
         * @throw An exception deriving from `armem::error::ArMemError` if the snapshot is missing.
         */
        virtual EntitySnapshot& getEntitySnapshot(const MemoryID& id)
        {
            return const_cast<EntitySnapshot&>(const_cast<const EntityStorage*>(this)->getEntitySnapshot(id));
        }
        virtual const EntitySnapshot& getEntitySnapshot(const MemoryID& id) const
        {
            const Entity& entity = getEntity(id);

            if (id.hasTimestamp())
            {
                return entity.getSnapshot(id);
            }
            else
            {
                return entity.getLatestSnapshot();
            }
        }


        /// Return `name`.
        virtual std::string getKeyString() const override
        {
            return name;
        }


    protected:

        /// Throw an `armem::error::StorageNameMismatch` if `expectedName != actualName`.
        static void checkStorageName(const std::string& expectedName, const std::string& actualName,
                                     const std::string& storageTerm)
        {
            if (expectedName != actualName)
            {
                throw armem::error::StorageNameMismatch(expectedName, storageTerm, actualName);
            }
        }

    public:

        /// The storage's name.
        std::string name;

    };


    /**
     * @brief An entity storage with a specific (Aron) type.
     */
    template <class ValueT>
    class TypedEntityStorage : public EntityStorage<ValueT>
    {
    public:

        TypedEntityStorage()
        {}
        TypedEntityStorage(const std::string& name, aron::type::AronTypePtr aronType = nullptr) :
            EntityStorage<ValueT>(name), aronType(aronType)
        {
        }

        virtual ~TypedEntityStorage() override
        {}


    public:

        /// The expected Aron type. May be nullptr, in which case no type information is available.
        aron::type::AronTypePtr aronType;

    };

}
