#include "EntityQueryProcessor.h"

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error/ArMemError.h"

#include "../core/ice_conversions.h"


namespace armarx::armem
{

    void EntityQueryProcessor::process(Entity& result, const query::EntityQuery& query, const Entity& entity) const
    {
        if (auto q = dynamic_cast<const query::entity::All*>(&query))
        {
            process(result, *q, entity);
        }
        else if (auto q = dynamic_cast<const query::entity::Single*>(&query))
        {
            process(result, *q, entity);
        }
        else if (auto q = dynamic_cast<const query::entity::Range*>(&query))
        {
            process(result, *q, entity);
        }
        else
        {
            throw armem::error::UnknownQueryType("entity snapshot", query);
        }
    }

    void EntityQueryProcessor::process(Entity& result, const query::entity::All& query, const Entity& entity) const
    {
        (void) query;
        // Copy this entitiy and its contents.
        result = entity;
    }

    void EntityQueryProcessor::process(Entity& result, const query::entity::Single& query, const Entity& entity) const
    {
        if (query.timestamp < 0)
        {
            try
            {
                result.addSnapshot(entity.getLatestSnapshot());
            }
            catch (const armem::error::EntityHistoryEmpty&)
            {
                // Leave empty.
            }
        }
        else
        {
            Time time = fromIce<Time>(query.timestamp);
            try
            {
                result.addSnapshot(entity.getSnapshot(time));
            }
            catch (const armem::error::MissingEntry&)
            {
                // Leave empty.
            }
        }
    }

    void EntityQueryProcessor::process(Entity& result, const query::entity::Range& query, const Entity& entity) const
    {
        if (query.minTimestamp <= query.maxTimestamp || query.minTimestamp < 0 || query.maxTimestamp < 0)
        {
            Time min = fromIce<Time>(query.minTimestamp);
            Time max = fromIce<Time>(query.maxTimestamp);
            process(result, min, max, entity, false);
        }
    }

    void EntityQueryProcessor::process(Entity& result, const Time& min, const Time& max, const Entity& entity, bool timestampsOnly) const
    {
        // Returns an iterator pointing to the first element that is not less than (i.e. greater or equal to) key.
        auto begin = min.toMicroSeconds() > 0 ? entity.history.lower_bound(min) : entity.history.begin();
        // Returns an iterator pointing to the first element that is *greater than* key.
        auto end = max.toMicroSeconds() > 0 ? entity.history.upper_bound(max) : entity.history.end();

        for (auto it = begin; it != end && it != entity.history.end(); ++it)
        {
            if (timestampsOnly)
            {
                result.history.emplace(it->second->time, nullptr);
            }
            else
            {
                result.addSnapshot(*it->second);
            }
        }
    }

    void EntityQueryProcessor::process(Entity& result, const query::entity::Timestamps& query, const Entity& entity) const
    {
        if (query.minTimestamp <= query.maxTimestamp || query.minTimestamp < 0 || query.maxTimestamp < 0)
        {
            Time min = fromIce<Time>(query.minTimestamp);
            Time max = fromIce<Time>(query.maxTimestamp);
            process(result, min, max, entity, true);
        }
    }

    data::EntityPtr EntityQueryProcessor::processToIce(const query::EntityQuery& query, const Entity& entity) const
    {
        return toIce<data::EntityPtr>(process(query, entity));
    }
}
