/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::armem
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::armem

#define ARMARX_BOOST_TEST

#include <RobotAPI/Test.h>
#include "../query/EntityQueryProcessor.h"
#include "../error.h"


#include <iostream>

#include <SimoxUtility/algorithm/get_map_keys_values.h>
#include <SimoxUtility/algorithm/string.h>


namespace armem = armarx::armem;
namespace aron = armarx::aron;


namespace ArMemQueryTest
{

    class UnknownEntityQuery : public armem::query::EntityQuery
    {
    };

    struct Fixture
    {
        armem::EntityPtr entity;

        armem::EntityQueryProcessor processor;
        armem::query::EntityQueryPtr entityQuery;

        std::vector<armem::EntityPtr> results;

        Fixture()
        {
            entity = std::make_unique<armem::Entity>("entity");

            armem::EntitySnapshot snapshot;
            snapshot.time = armem::Time::microSeconds(1000);
            entity->addSnapshot(snapshot);
            snapshot.time = armem::Time::microSeconds(2000);
            entity->addSnapshot(snapshot);
            snapshot.time = armem::Time::microSeconds(3000);
            entity->addSnapshot(snapshot);
            snapshot.time = armem::Time::microSeconds(4000);
            entity->addSnapshot(snapshot);
            snapshot.time = armem::Time::microSeconds(5000);
            entity->addSnapshot(snapshot);
        }
        ~Fixture()
        {

        }

        template <class QueryT>
        void addResults(QueryT query = {})
        {
            entityQuery = new QueryT(query);

            results.emplace_back(processor.process(query, *entity));
            results.emplace_back(processor.process(*entityQuery, *entity));
        }
    };

}


BOOST_FIXTURE_TEST_SUITE(ArMemQueryTest, ArMemQueryTest::Fixture)


BOOST_AUTO_TEST_CASE(test_entity_Single_latest)
{
    addResults(armem::query::entity::Single());
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), 1);
        BOOST_CHECK_EQUAL(result->history.begin()->second->time, entity->getLatestSnapshot().time);
        BOOST_CHECK_NE(result->history.begin()->second.get(), entity->history.rbegin()->second.get());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Single_existing)
{
    addResults(armem::query::entity::Single { 3000 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), 1);
        BOOST_CHECK_EQUAL(result->history.begin()->second->time, armem::Time::microSeconds(3000));
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Single_non_existing)
{
    addResults(armem::query::entity::Single { 3500 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), 0);
    }
}


BOOST_AUTO_TEST_CASE(test_entity_All)
{
    addResults(armem::query::entity::All());
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), entity->history.size());
        for (auto it = result->history.begin(), jt = entity->history.begin();
             it != result->history.end(); ++it, ++jt)
        {
            BOOST_CHECK_EQUAL(it->first, jt->first);
            BOOST_CHECK_EQUAL(it->second->time, jt->second->time);
            BOOST_CHECK_NE(it->second.get(), jt->second.get());
        }
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Range_slice)
{
    addResults(armem::query::entity::Range{ 1500, 3500 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), 2);

        std::vector<armem::Time> times = simox::get_keys(result->history);
        std::vector<armem::Time> expected {
            armem::Time::microSeconds(2000), armem::Time::microSeconds(3000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Range_exact)
{
    addResults(armem::query::entity::Range{ 2000, 4000 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), 3);

        std::vector<armem::Time> times = simox::get_keys(result->history);
        std::vector<armem::Time> expected {
            armem::Time::microSeconds(2000), armem::Time::microSeconds(3000), armem::Time::microSeconds(4000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}

BOOST_AUTO_TEST_CASE(test_entity_Range_all)
{
    addResults(armem::query::entity::Range{ 0, 10000 });
    addResults(armem::query::entity::Range{ -1, -1 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), entity->history.size());

        std::vector<armem::Time> times = simox::get_keys(result->history);
        std::vector<armem::Time> expected = simox::get_keys(entity->history);
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Range_empty)
{
    addResults(armem::query::entity::Range{ 2400, 2600 });
    addResults(armem::query::entity::Range{ 6000, 1000 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), 0);
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Range_from_start)
{
    addResults(armem::query::entity::Range{ -1, 2500 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), 2);

        std::vector<armem::Time> times = simox::get_keys(result->history);
        std::vector<armem::Time> expected {
            armem::Time::microSeconds(1000), armem::Time::microSeconds(2000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_Range_to_end)
{
    addResults(armem::query::entity::Range{ 2500, -1 });
    BOOST_REQUIRE_GT(results.size(), 0);

    for (const auto& result : results)
    {
        BOOST_CHECK_EQUAL(result->name, entity->name);
        BOOST_CHECK_EQUAL(result->history.size(), 3);

        std::vector<armem::Time> times = simox::get_keys(result->history);
        std::vector<armem::Time> expected {
            armem::Time::microSeconds(3000), armem::Time::microSeconds(4000), armem::Time::microSeconds(5000)
        };
        BOOST_CHECK_EQUAL_COLLECTIONS(times.begin(), times.end(), expected.begin(), expected.end());
    }
}


BOOST_AUTO_TEST_CASE(test_entity_UnknownQueryType)
{
    BOOST_CHECK_THROW(processor.process(UnknownEntityQuery(), *entity), armem::error::UnknownQueryType);
    std::string msg;
    try
    {
        processor.process(UnknownEntityQuery(), *entity);
    }
    catch (const armem::error::UnknownQueryType& e)
    {
        msg = e.what();
    }
    BOOST_TEST_MESSAGE(msg);
    BOOST_CHECK(simox::alg::contains(msg, "UnknownEntityQuery"));
}


BOOST_AUTO_TEST_SUITE_END()
