/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "AronDataNavigatorReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/aroncore/AronException.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronListDataNavigator.h>
#include <RobotAPI/libraries/aron/aroncore/navigators/datanavigator/AronPrimitiveDataNavigator.h>

using namespace std;

namespace armarx
{
    namespace aron
    {
        namespace io
        {
            AronDataNavigatorReader::AronDataNavigatorReader(const data::AronDataPtr& aron) :
                AronDataClassReader<data::AronDataPtr>(aron),
                input(datanavigator::AronDataNavigator::FromAronData(aron, AronPath()))
            {

            }

            datanavigator::AronDataNavigatorPtr AronDataNavigatorReader::getChildOfLatestAronDataNavigator()
            {
                if (!readInitialStart)
                {
                    readInitialStart = true;
                    return input;
                }
                AronDataNavigatorReaderTokenPtr lastToken = stack.top();
                return lastToken->getElementAndIncreaseCnt();
            }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    int AronDataNavigatorReader::readStart##upperType() \
    { \
        datanavigator::AronDataNavigatorPtr current_nav = getChildOfLatestAronDataNavigator(); \
        int c = current_nav->childrenSize(); \
        AronDataNavigatorReaderTokenPtr newToken = AronDataNavigatorReaderTokenPtr(new AronDataNavigatorReaderToken(current_nav)); \
        stack.push(newToken); \
        return c; \
    } \
    \
    bool AronDataNavigatorReader::readEnd##upperType() \
    { \
        AronDataNavigatorReaderTokenPtr token = stack.top(); \
        auto desc = token->getDescriptor(); \
        if (desc != AronDataDescriptor::eAron##upperType) \
        { \
            throw exception::AronDataDescriptorNotValidException("AronDataNavigatorReader", "readEnd" + std::string(#upperType), "A token in the stack has the wrong type. Expected " + std::string(#upperType), desc); \
            return false; \
        } \
        \
        if (token->checkedAllChildren()) \
        { \
            stack.pop(); \
            return true; \
        } \
        return false; \
    }

            HANDLE_CONTAINER_DATA
#undef RUN_ARON_MACRO

            // Complex Data
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    std::vector<int> AronDataNavigatorReader::readStart##upperType() \
    { \
        datanavigator::AronDataNavigatorPtr nav = getChildOfLatestAronDataNavigator(); \
        datanavigator::AronNDArrayDataNavigatorPtr casted = datanavigator::AronNDArrayDataNavigator::DynamicCast(nav); \
        std::vector<int> dims = casted->getDimensions(); \
        return dims; \
    } \
    \
    bool AronDataNavigatorReader::readEnd##upperType(unsigned char* data) \
    { \
        if(data == NULL) \
        { \
            throw exception::AronException("AronDataNavigatorReader", "readEnd" + std::string(#upperType), "The corresponding data pointer of an complex aron ptr is NULL."); \
        } \
        datanavigator::AronDataNavigatorPtr nav = getChildOfLatestAronDataNavigator(); \
        datanavigator::AronNDArrayDataNavigatorPtr casted = datanavigator::AronNDArrayDataNavigator::DynamicCast(nav); \
        std::vector<int> dims = casted->getDimensions(); \
        memcpy(data, casted->getData(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>())); \
        return true; \
    }

            HANDLE_COMPLEX_DATA
#undef RUN_ARON_MACRO

            // Read primitives
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    void AronDataNavigatorReader::readPrimitive(lowerType& t) \
    { \
        AronDataNavigatorReaderTokenPtr token = stack.top(); \
        datanavigator::AronDataNavigatorPtr nav = token->getElementAndIncreaseCnt(); \
        datanavigator::Aron##upperType##DataNavigatorPtr casted = datanavigator::Aron##upperType##DataNavigator::DynamicCast(nav); \
        t = casted->getValue(); \
    }

            HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO

            // ReadKey
            std::string AronDataNavigatorReader::readKey()
            {
                AronDataNavigatorReaderTokenPtr token = stack.top();
                return token->getCurrentKey();
            }

            bool AronDataNavigatorReader::readKey(const std::string& k)
            {
                AronDataNavigatorReaderTokenPtr token = stack.top();
                token->setCurrentKey(k);
                return true;
            }
        }
    }
}
