/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller (fabian dot peller at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <numeric>

#include "AronDataJSONWriter.h"

namespace armarx
{
    namespace aron
    {
        namespace io
        {
            bool AronDataJSONWriter::writeStartDict()
            {
                if (openedTagsStack.size() != 0)
                {
                    if (openedTagsStack.top().first == eOpenList && openedTagsStack.top().second > 0)
                    {
                        writeNewLine();
                    }
                    openedTagsStack.top().second++;
                }

                openedTagsStack.push(std::make_pair(eOpenDict, 0));
                ss << "{";
                writeNewLineWithIndentIncrease();
                return true;
            }

            bool AronDataJSONWriter::writeEndDict()
            {
                if (openedTagsStack.size() == 0)
                {
                    throw exception::AronException("AronDataJSONWriter", "writeEndDict", "called method without calling writeStartDict first. The stack is empty.");
                }

                if (openedTagsStack.top().first != eOpenDict)
                {
                    throw exception::AronException("AronDataJSONWriter", "writeEndDict", "called method without calling writeStartDict first. Wrong type in stack.");
                }

                if (openedTagsStack.top().second > 0)
                {
                    removeLastOccurenceOf(",");
                }

                openedTagsStack.pop();
                writeNewLineWithIndentDecrease();
                ss << "}";

                if (openedTagsStack.size() != 0)
                {
                    ss << ",";
                }
                return true;
            }

            bool AronDataJSONWriter::writeStartList()
            {
                if (openedTagsStack.size() != 0)
                {
                    openedTagsStack.top().second++;
                }

                openedTagsStack.push(std::make_pair(eOpenList, 0));
                ss << "[";
                writeNewLineWithIndentIncrease();
                return true;
            }

            bool AronDataJSONWriter::writeEndList()
            {
                if (openedTagsStack.size() == 0)
                {
                    throw exception::AronException("AronDataJSONWriter", "writeEndList", "called method without calling writeStartList first. The stack is empty.");
                }

                if (openedTagsStack.top().first != eOpenList)
                {
                    throw exception::AronException("AronDataJSONWriter", "writeEndList", "called method without calling writeStartList first. Wrong type in stack.");
                }

                if (openedTagsStack.top().second == 0)
                {
                    trimLastWhitespaces();
                    decreaseIndent();
                }
                else
                {
                    removeLastOccurenceOf(",");
                    writeNewLineWithIndentDecrease();
                }

                openedTagsStack.pop();
                ss << "]";

                if (openedTagsStack.size() != 0)
                {
                    ss << ",";
                }
                return true;
            }

            bool AronDataJSONWriter::writeNDArray(const std::vector<int>& dims, const unsigned char* data)
            {
                openedTagsStack.top().second++;
                ss << "[ ";
                int size = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
                for (int i = 0; i < size; ++i)
                {
                    ss << +data[i];
                    if (i != (size - 1))
                    {
                        ss << ", ";
                    }
                }
                ss << " ],";
                return true;
            }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    bool AronDataJSONWriter::writePrimitive(const lowerType& x) \
    { \
        openedTagsStack.top().second++; \
        ss << "\"" << x << #lowerType[0] << "\", "; \
        return true; \
    }

            HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO

            bool AronDataJSONWriter::writeKey(const std::string& k)
            {
                if (openedTagsStack.size() == 0)
                {
                    throw exception::AronException("AronDataJSONWriter", "writeKey", "called method without calling writeStartDict first. The stack is empty.");
                }
                if (openedTagsStack.top().first != eOpenDict)
                {
                    throw exception::AronException("AronDataJSONWriter", "writeKey", "called method without calling writeStartDict first. Wrong type in stack.");
                }
                if (openedTagsStack.top().second > 0)
                {
                    writeNewLine();
                }

                ss << "\"" << k << "\"" << ": ";
                return true;
            }
        }
    }
}
