/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller (fabian dot peller at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <numeric>

#include "AronDataXMLWriter.h"

namespace armarx
{
    namespace aron
    {
        namespace io
        {
            bool AronDataXMLWriter::writeStartDict()
            {
                if (openedTagsStack.size() != 0)
                {
                    openedTagsStack.top().second++;
                }

                openedTagsStack.push(std::make_pair(eOpenDict, 0));
                ss << "<Dict>";
                writeNewLineWithIndentIncrease();
                return true;
            }

            bool AronDataXMLWriter::writeEndDict()
            {
                if (openedTagsStack.size() == 0)
                {
                    throw exception::AronException("AronDataXMLWriter", "writeEndDict", "called method without calling writeStartDict first. The stack is empty.");
                }

                if (openedTagsStack.top().first != eOpenDict)
                {
                    throw exception::AronException("AronDataXMLWriter", "writeEndDict", "called method without calling writeStartDict first. Wrong type in stack.");
                }

                openedTagsStack.pop();
                trimLastWhitespaces();
                writeNewLineWithIndentDecrease();
                ss << "</Dict>";
                writeNewLine();
                return true;
            }

            bool AronDataXMLWriter::writeStartList()
            {
                if (openedTagsStack.size() != 0)
                {
                    openedTagsStack.top().second++;
                }

                openedTagsStack.push(std::make_pair(eOpenList, 0));
                ss << "<List>";
                writeNewLineWithIndentIncrease();
                return true;
            }

            bool AronDataXMLWriter::writeEndList()
            {
                if (openedTagsStack.size() == 0)
                {
                    throw exception::AronException("AronDataXMLWriter", "writeEndList", "called method without calling writeStartList first. The stack is empty.");
                }

                if (openedTagsStack.top().first != eOpenList)
                {
                    throw exception::AronException("AronDataXMLWriter", "writeEndList", "called method without calling writeStartList first. Wrong type in stack.");
                }

                trimLastWhitespaces();
                if (openedTagsStack.top().second == 0)
                {
                    decreaseIndent();
                }
                else
                {
                    writeNewLineWithIndentDecrease();
                }

                openedTagsStack.pop();
                ss << "</List>";
                writeNewLine();
                return true;
            }

            bool AronDataXMLWriter::writeNDArray(const std::vector<int>& dims, const unsigned char* data)
            {
                openedTagsStack.top().second++;
                ss << "<NDArray>";
                int size = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
                for (int i = 0; i < size; ++i)
                {
                    ss << +data[i];
                    if (i != (size-1))
                    {
                        ss << ", ";
                    }
                }
                ss << "</NDArray>";
                writeNewLine();
                return true;
            }

#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    bool AronDataXMLWriter::writePrimitive(const lowerType& x) \
    { \
        openedTagsStack.top().second++; \
        ss << ("<" + std::string(#upperType) + ">") << x << ("</" + std::string(#upperType) + ">"); \
        writeNewLine(); \
        return true; \
    }

            HANDLE_PRIMITIVE_TYPES
#undef RUN_ARON_MACRO

            bool AronDataXMLWriter::writeKey(const std::string& k)
            {
                if (openedTagsStack.size() == 0)
                {
                    throw exception::AronException("AronDataXMLWriter", "writeKey", "called method without calling writeStartDict first. The stack is empty.");
                }
                if (openedTagsStack.top().first != eOpenDict)
                {
                    throw exception::AronException("AronDataXMLWriter", "writeKey", "called method without calling writeStartDict first. Wrong type in stack.");
                }
                if (openedTagsStack.top().second > 0)
                {
                    //writeNewLine();
                }

                ss << "<Key>" << k << "</Key>";
                writeNewLine();
                return true;
            }
        }
    }
}
